/**
 * Create a `StarryNight` that can highlight things based on the given
 * `grammars`.
 * This is async to facilitate async loading and registering, which is currently
 * only used for WASM.
 *
 * @param {Array<Grammar>} grammars
 *   Grammars to support.
 * @param {Options | null | undefined} [options]
 *   Configuration (optional).
 * @returns
 *   Promise that resolves to an instance which highlights based on the bound
 *   grammars.
 */
export function createStarryNight(grammars: Array<Grammar>, options?: Options | null | undefined): Promise<{
    flagToScope: (flag: string) => string | undefined;
    highlight: (value: string, scope: string) => Root;
    missingScopes: () => Array<string>;
    register: (grammars: Array<Grammar>) => Promise<undefined>;
    scopes: () => Array<string>;
}>;
export type Root = import('hast').Root;
/**
 * TextMate grammar with some extra info.
 * List of extensions that only match if used w/ a dot (example: `[]`).
 */
export type Grammar = {
    /**
     *   Scope (example: `source.mdx`).
     */
    scopeName: string;
    /**
     *   List of names (example: `['mdx']`).
     */
    names: Array<string>;
    /**
     * List of scopes that are needed for this grammar to work (example:
     * `['source.tsx']`).
     */
    dependencies?: Array<string>;
    /**
     *   List of extensions (example: `['.mdx']`).
     */
    extensions: Array<string>;
    extensionsWithDot?: Array<string>;
    /**
     *   TextMate patterns.
     */
    patterns: Array<Rule>;
    /**
     * TextMate repository.
     */
    repository?: Record<string, Rule>;
    /**
     * TextMate injections.
     */
    injections?: Record<string, Rule>;
};
export type Rule = RuleDefinition | RuleInclude | RuleName;
export type RuleName = {
    name: string;
    include?: never;
    begin?: never;
    match?: never;
};
export type RuleInclude = {
    include: string;
    name?: string;
    begin?: string;
    end?: string;
    match?: string;
};
export type RuleDefinition = {
    match?: string;
    captures?: Captures;
    begin?: string;
    beginCaptures?: Captures;
    end?: string;
    endCaptures?: Captures;
    while?: string;
    whileCaptures?: Captures;
    name?: string;
    contentName?: string;
    patterns?: Array<Rule>;
    repository?: Record<string, Rule>;
    injections?: Record<string, Rule>;
    applyEndPatternLast?: boolean;
};
export type Captures = Record<string, Rule>;
/**
 * Configuration (optional).
 */
export type Options = {
    /**
     * Get a URL to the oniguruma WASM, typically used in browsers.
     */
    getOnigurumaUrlFetch?: GetOnigurumaUrl | null | undefined;
    /**
     * Get a URL to the oniguruma WASM, typically used in Node.js.
     */
    getOnigurumaUrlFs?: GetOnigurumaUrl | null | undefined;
};
/**
 * Get a URL to the oniguruma WASM.
 *
 * > 👉 **Note**: this must currently result in a version 1 URL of
 * > `onig.wasm` from `vscode-oniguruma`.
 *
 * > ⚠️ **Danger**: when you use this functionality, your project might break at
 * > any time (when reinstalling dependencies), except when you make sure that
 * > the WASM binary you load manually is what our internally used
 * > `vscode-oniguruma` dependency expects.
 * > To solve this, you could for example use an npm script called `dependencies`
 * > (which runs everytime `node_modules` is changed) which copies
 * > `vscode-oniguruma/release/onig.wasm` to the place you want to host it.
 */
export type GetOnigurumaUrl = () => Promise<URL> | URL;
